"""Utility Functions"""

import os
import rsa
import sys
import socket
from os.path import exists, basename
from textwrap import TextWrapper
from itertools import chain
from smtplib import SMTP_SSL as SMTP
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.encoders import encode_base64
from subprocess import Popen, PIPE
from promus.command import error
import promus.core.git as git
from promus.core.ssh import get_keys


def is_exe(fpath):
    """Checks to see if the specified path is an executable. """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def external_executables(ext_exec):
    """Checks to see if your system has access to the executables
    specified by `ext_exec`. It returns a list of the missing and
    found executables. """
    missing = []
    found = []
    for program in ext_exec:
        fpath, _ = os.path.split(program)
        if fpath:
            if is_exe(program):
                found.append(program)
                continue
        else:
            in_path = False
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    found.append(exe_file)
                    in_path = True
                    break
            if in_path:
                continue
        missing.append(program)
    return missing, found


def check_promus_dependencies():
    """This function will terminate the program if it finds that some
    external executables are missing. At the moment there are no
    alternatives as to how to account for their absence. """
    missing, _ = external_executables(['git', 'ssh-keygen'])
    if missing:
        msg = "ERROR: your system is missing the following executables: \n" \
              "    %s \n" \
              "Install them or make sure that they are in your `PATH`.\n"
        error(msg % missing)


def exec_cmd(cmd, verbose=False):
    "Run a subprocess and return its output and errors. "
    if verbose:
        out = sys.stdout
        err = sys.stderr
    else:
        out = PIPE
        err = PIPE
    process = Popen(cmd, shell=True, stdout=out, stderr=err)
    out, err = process.communicate()
    return out, err, process.returncode


def user_input(prompt, default):
    "Get an input given a default value. "
    if default != '':
        newval = raw_input("%s [%s]: " % (prompt, default))
        if newval == '':
            newval = default
    else:
        newval = default
        while newval == '':
            newval = raw_input("%s: " % prompt)
    return newval.strip()


def encrypt_to_file(msg, fname, keyfile):
    """Encrypt msg to file `fname` using the key given by the path
    `keyfile`."""
    with open(keyfile, 'r') as keyfp:
        keydata = keyfp.read()
    key = rsa.PrivateKey.load_pkcs1(keydata)
    msg = rsa.encrypt(msg, key)
    with open(fname, 'w') as msgfp:
        msgfp.write(msg)


def decrypt_from_file(fname, keyfile):
    """Decrypt a message in the file `fname` using the key given by
    the path `keyfile`"""
    with open(keyfile, 'r') as keyfp:
        keydata = keyfp.read()
    key = rsa.PrivateKey.load_pkcs1(keydata)
    with open(fname, 'r') as msgfp:
        msg = msgfp.read()
    return rsa.decrypt(msg, key)


def wrap_msg(msg, width=70, tab=1):
    "wraps the msg to the specified width and tab indentation. "
    width -= tab*4
    wrapper = TextWrapper(width=width, break_long_words=False)
    tab = '\n%s' % ('    '*tab)
    return tab.join(wrapper.wrap(msg))


def make_dir(path):
    "Create a directory if it does not exist and return True. "
    if exists(path):
        return False
    os.makedirs(path)
    return True


def parse_list(line, delim=','):
    "Return a list of strings separated by delim. "
    return [item.strip() for item in line.split(delim) if item.strip() != '']


def tokenizer(string, delim):
    "Return an iterator for `string.split(delim)`. "
    caret = 0
    while True:
        try:
            end = string.index(delim, caret)
        except ValueError:
            yield string[caret:]
            return
        yield string[caret:end]
        caret = end + 1


def merge_lines(str1, str2):
    """The inputs need to be strings separated by the newline character.
    It will return a set containing all the lines. """
    return set(chain(tokenizer(str1, '\n'), tokenizer(str2, '\n')))


def strip(string):
    "Wrapper around the strip function. "
    if string:
        return string.strip()
    return None


def send_mail(send_to, subject, text, html, files=None):
    """Send an email. `send_to` must be a list of email address to
    which the email will be sent. You can email a `subject` as well
    as two versions of the email: text and html. You may optionally
    attach files by providing a list of them. """
    if not send_to:
        return
    msg = MIMEMultipart('alternative')
    msg['Subject'] = subject
    msg['From'] = 'promus@%s' % socket.gethostname()
    msg['To'] = ','.join(send_to)

    msg.attach(MIMEText(text, 'plain'))
    htmlmsg = MIMEMultipart()
    htmlmsg.attach(MIMEText(html, 'html'))

    if files is None:
        files = []
    for file_ in files:
        part = MIMEBase('application', "octet-stream")
        part.set_payload(open(file_, "rb").read())
        encode_base64(part)
        part.add_header('Content-Disposition',
                        'attachment; filename="%s"' % basename(file_))
        htmlmsg.attach(part)
    msg.attach(htmlmsg)
    server = git.config('host.smtpserver')
    conn = SMTP(server)
    conn.set_debuglevel(False)
    id_key, _ = get_keys()
    passfile = '%s/.promus/password.pass' % os.environ['HOME']
    password = decrypt_from_file(passfile, id_key)
    if password:
        conn.login(git.config('host.username'), password)
    try:
        conn.sendmail(git.config('host.email'), send_to, msg.as_string())
    finally:
        conn.close()



"""Command

Collection of functions to create the command line utility.

"""

import sys
from dateutil import parser
from datetime import datetime
from subprocess import Popen, PIPE


def error(msg):
    "Print a message to the standard error stream and exit. "
    sys.stderr.write(msg)
    sys.exit(2)


def warn(msg):
    "Print a message to the standard error "
    sys.stderr.write(msg)


def import_mod(name):
    "Return a module by string. "
    mod = __import__(name)
    for sub in name.split(".")[1:]:
        mod = getattr(mod, sub)
    return mod


def exec_cmd(cmd, verbose=False):
    "Run a subprocess and return its output and errors. "
    if verbose:
        out = sys.stdout
        err = sys.stderr
    else:
        out = PIPE
        err = PIPE
    process = Popen(cmd, shell=True, stdout=out, stderr=err)
    out, err = process.communicate()
    return out, err, process.returncode


# pylint: disable=E1103
# The reason for this pylint error is that it thinks of
# `now` as a `tuple` instead of a `datetime` object.
def date(short=False):
    "Return the current date as a string. "
    if isinstance(short, str):
        now = parser.parse(str(short))
        return now.strftime("%a %b %d, %Y %r")
    now = datetime.now()
    if not short:
        return now.strftime("%a %b %d, %Y %r")
    return now.strftime("%Y-%m-%d-%H-%M-%S")



"""promus version

version_info conforms to PEP 386

(major, minor, micro, alpha/beta/rc/final, #)
(1, 1, 2, 'alpha', 0) => "1.1.2.dev"
(1, 2, 0, 'beta', 2) => "1.2b2"

"""

VERSION_INFO = (0, 7, 0, 'beta', 2)


def get_version():
    """Return a PEP-386 compliant version number from version_info."""
    assert len(VERSION_INFO) == 5
    assert VERSION_INFO[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if VERSION_INFO[2] == 0 else 3
    main = '.'.join([str(part) for part in VERSION_INFO[:parts]])

    sub = ''
    if VERSION_INFO[3] == 'alpha' and VERSION_INFO[4] == 0:
        sub = '.dev'
    elif VERSION_INFO[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[VERSION_INFO[3]] + str(VERSION_INFO[4])

    return str(main + sub)

VERSION = get_version()



"""Add

Adds a user if promus is acting as a host or adds a host if promus is
acting as a client.

"""

import os
import sys
import socket
import textwrap
import promus.core.ssh as ssh
import promus.core.git as git
import promus.core.util as util

DESC = """
adds a host to your ssh configuration file and sends your public git
key.

"""


def add_parser(subp, raw):
    "Add a parser to the main subparser. "
    tmpp = subp.add_parser('add', help='add a host',
                           formatter_class=raw,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('type', type=str, metavar='TYPE',
                      choices=['host', 'user'],
                      help='One of the following: host')
    tmpp.add_argument('host', type=str,
                      help='the name of the private key sent to you')


def add_host(arg):
    """Display your public key and your git key. """
    disp = sys.stdout.write
    try:
        util.exec_cmd('chmod 700 %s' % arg.host, True)
    except OSError:
        disp('ERROR: Private key `%s` not found\n' % arg.host)
        return
    pub_key = ssh.get_public_key(arg.host)
    print pub_key
    _, gitkey = ssh.get_keys()
    gitkey = ssh.get_public_key(gitkey)
    host = socket.gethostname()
    email = git.config('user.email')
    master = os.environ['USER']
    master_name = git.config('user.name')
    alias = git.config('host.alias')
    cmd = 'ssh -i {host} {host} ' \
          '{pub},{gitkey},{email},{master},{name},{hostname},{alias}'
    cmd = cmd.format(host=arg.host, gitkey=gitkey, email=email,
                     master=master, name=master_name, hostname=host,
                     alias=alias, pub=pub_key[-20:])
    util.exec_cmd(cmd, True)
    os.remove(arg.host)
    config = ssh.read_config()
    found = False
    for entry in config:
        if arg.host.replace('@', '-') in entry.split():
            found = True
            disp('Existing entry: `Host %s`\n' % entry)
            break
    if not found:
        _, gitkey = ssh.get_keys()
        entry = '%s' % arg.host.replace('@', '-')
        config[entry] = dict()
        user, host = arg.host.split('@')
        config[entry]['HostName'] = host
        config[entry]['User'] = user
        config[entry]['IdentityFile'] = gitkey
        ssh.write_config(config)


EMAIL_TXT = """Hello {name},

You public key has been added and you may now connect to {host} as
{user} using your public key. You may only run git commands however.

- Promus

"""
EMAIL_HTML = """<p>Hello {name},</p>
<p>You public key has been added and you may now connect to {host} as
{user} using your public key. You may only run git commands however.
</p>
<p>
<strong>- Promus</strong>
</p>
"""


def add_user(_):
    """Display all the users that have access to your account. """
    # useremail = arg.host
    info = os.environ['SSH_ORIGINAL_COMMAND']
    pub, key, email, user, username, host, alias = info.split(',')
    users, pending, unknown = ssh.read_authorized_keys()
    # Remove access from private key
    for entry in pending:
        if entry[-20:] == pub:
            pub = entry
            break
    del pending[pub]
    if email not in users:
        users[email] = dict()
    key_type, key_val = key.split()
    content = [user,
               username,
               alias,
               key_type,
               '%s@%s' % (user, host)]
    users[email][key_val] = content
    ssh.write_authorized_keys(users, pending, unknown)
    # if useremail != email:
    # Needs to modify message if the emails do not match
    util.send_mail([email, git.config('host.email')],
                   'Connection successful',
                   EMAIL_TXT.format(name=username,
                                    host=socket.gethostname(),
                                    user=os.environ['USER']),
                   EMAIL_HTML.format(name=username,
                                     host=socket.gethostname(),
                                     user=os.environ['USER']))


def run(arg):
    """Run command. """
    func = {
        'host': add_host,
        'user': add_user,
    }
    func[arg.type](arg)



"""promus setup script"""

import imp
import os.path as pt
from setuptools import setup


def get_version():
    "Get version & version_info without importing promus.__init__ "
    path = pt.join(pt.dirname(__file__), 'promus', '__version__.py')
    mod = imp.load_source('promus_version', path)
    return mod.VERSION, mod.VERSION_INFO

VERSION, VERSION_INFO = get_version()

DESCRIPTION = "Remote manager to handle git commands."
LONG_DESCRIPTION = """
Promus is a remote manager designed to create and manage `git`
repositories in a remote server without the need of administrator
privileges.
"""

DEV_STATUS_MAP = {
    'alpha': '3 - Alpha',
    'beta': '4 - Beta',
    'rc': '4 - Beta',
    'final': '5 - Production/Stable'
}
if VERSION_INFO[3] == 'alpha' and VERSION_INFO[4] == 0:
    DEVSTATUS = '2 - Pre-Alpha'
else:
    DEVSTATUS = DEV_STATUS_MAP[VERSION_INFO[3]]


setup(name='promus',
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      keywords='ssh git transfer files collaborate access',
      author='Manuel Lopez',
      author_email='jmlopez.rod@gmail.com',
      url='http://promus.readthedocs.org',
      license='BSD License',
      packages=[
          'promus',
          'promus.command',
          'promus.core',
          ],
      platforms=['Darwin', 'Linux'],
      scripts=['bin/promus'],
      install_requires=[
          'pysftp>=0.2.8',
          'rsa>=3.1.4',
          ],
      package_data={'promus.git': ['gitignore'],
                    'promus.paster': ['latex/*', 'python/*'],
                    },
      include_package_data=True,
      classifiers=['Development Status :: %s' % DEVSTATUS,
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: MacOS :: MacOS X',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2.7',
                   'Topic :: Communications :: File Sharing',
                   'Topic :: Software Development :: Version Control',
                   ],
      )
